### esp32_ble
import ubluetooth
from micropython import const
from binascii import hexlify, unhexlify
from time import sleep_ms, ticks_ms, ticks_diff

class ESP32_BLE():
    
    mac_addr = 'bluetooth mac-address'
    peripheral_devices = 'connected peripheral device list'
    peripheral_connect_try = 'peripheral connect try count'
    peripheral_connected_count = 'peripheral connected count'
    peripheral_connect_failed_count = 'peripheral failed connection attempt'
    peripheral_disconnected_count = 'peripheral disconnected count'
    last_peripheral_addr = 'last connected peripheral address'
    
    def __init__(self,name=''):
        self.ble = ubluetooth.BLE()
        self.ble.active(True)
        self.mac_addr = str(hexlify((self.ble.config('mac'))[1]),'utf-8')
        
        if not name:
            self.name = "ESP32" + "-" + str(self.mac_addr, 'utf-8')[8:12]
        else:
            self.name = name
        
        self.ble.config(gap_name=name)
        
        self.ble.irq(self.ble_irq)

        #const
        _IRQ_PERIPHERAL_CONNECT = const(7)
        _IRQ_PERIPHERAL_DISCONNECT = const(8)
        
        self.peripheral_devices = []
        self.peripheral_connect_try = 0
        self.peripheral_connected_count = 0
        self.peripheral_connect_failed_count = 0
        self.peripheral_disconnected_count = 0
        self.last_peripheral_addr = None
        
    def ble_irq(self,event, data):
        if event is 7: #_IRQ_PERIPHERAL_CONNECT:
            conn_handle, addr_type, addr = data
            addr = hexlify(addr)
            self.peripheral_devices.append((conn_handle, addr, addr_type))
            self.peripheral_connected_count += 1
            self.last_peripheral_addr = {addr: addr_type}
            try:
                type(ble_callback.ble_peripheral_connected)
            except:
                print(f"ble_peripheral_connected!!! {conn_handle} {addr} {addr_type}")
            else:
                ble_callback.ble_peripheral_connected(conn_handle, addr, addr_type)
        elif event is 8: #_IRQ_PERIPHERAL_DISCONNECT:
            conn_handle, addr_type, addr = data
            addr = hexlify(addr)
            if addr != b'000000000000':
                self.peripheral_devices.pop(self.peripheral_devices.index((conn_handle, addr, addr_type)))
                self.peripheral_disconnected_count += 1
            else:
                self.peripheral_connect_failed_count += 1
            try:
                type(ble_callback.ble_peripheral_disconnected)
            except:
                print(f"ble_peripheral_disconnected!!! {conn_handle} {addr} {addr_type}")
            else:
                ble_callback.ble_peripheral_disconnected(conn_handle, addr, addr_type)
            
    def get_conn_handle(self,mac_addr):
        dict = {}
        for element in self.peripheral_devices:
            dict[element[1]] = element[0]
        if mac_addr in dict:
            return dict[mac_addr]
        else:
            return None
        
    def get_mac_addr(self,con_handle):
        dict = {}
        for element in self.peripheral_devices:
            dict[element[0]] = element[1]
        if con_handle in dict:
            return dict[con_handle]
        else:
            return None
            
    def disconnect(self,obj): #obj is mac_addr or con_handle
        if type(obj) is not int:
            conn_handle = self.get_conn_handle(obj)
        else:
            conn_handle = obj
        all_handle = [handle[0] for handle in self.peripheral_devices]
        if conn_handle in all_handle:
            self.ble.gap_disconnect(conn_handle)
        else:
            print(f"connection handle {handle} not found!!!")
        
    def disconnect_all_devices(self):
        if len(self.peripheral_devices) is not 0:
            for devices in self.peripheral_devices:
                self.disconnect(devices[0])
                sleep_ms(200)    
### esp32_ble
### esp32_ble_callback
class ESP32_BLE_CALLBACK():
    def ble_peripheral_connected(self,conn_handle, addr, addr_type):
        #print(f"{addr} connected!")
        ble.peripheral_connection_TSF.set()
        
    def ble_peripheral_disconnected(self,conn_handle, addr, addr_type):
        if addr != b'000000000000':
            pass
            #print(f"{addr} disconnected!")
        else:
            #print('ble_peripheral_connect_failed!!!')
            ble.peripheral_connection_TSF.set()
### esp32_ble_callback
            
import uasyncio
from random import randint

ble = ESP32_BLE()
ble.peripheral_connection_TSF = uasyncio.ThreadSafeFlag()
ble_callback = ESP32_BLE_CALLBACK()
mac_addr = b'f3f06b581563'

def connect(connect_delay,connect_min,connect_max):
    ble.ble.gap_connect(1, unhexlify(mac_addr), connect_delay, connect_min, connect_max)
    
async def main_loop():
    connect_min = 99999
    connect_max = 0
    connect_list = []
    gap_connect_min = 8000
    gap_connect_max = 32000
    gap_connect_delay = 2000
    gap_connect_min = 0
    gap_connect_max = 0
    ct = 0
    cd = 0
    cf = 0
    while True:
        if len(connect_list) > 100:
            connect_list = [connect_mean]
        ct += 1
        connect_start = ticks_ms()
        try:
            connect(gap_connect_delay,gap_connect_min,gap_connect_max)
        except:
            sleep_ms(500)
        else:
            await ble.peripheral_connection_TSF.wait()
            if ble.peripheral_devices:
                cd += 1
                connect_time = ticks_ms() - connect_start
                connect_list.append(connect_time)
                if connect_time < connect_min:
                    connect_min = connect_time
                if connect_time > connect_max:
                    connect_max = connect_time
                connect_mean = int(sum(connect_list) / len(connect_list))
                print(f"connection time : {connect_time}")
                print(f"min: {connect_min} mean: {connect_mean} max: {connect_max} connect delay: {gap_connect_delay}")
                #print(connect_list)
                sleep_ms(500)
                if ble.peripheral_devices:
                   ble.disconnect_all_devices()
                random_sleep = randint(2000,8000)
                #print(f"random sleep:{random_sleep}")
            else:
                cf +=1
                gap_connect_delay += 1000
        print(f"ct: {ct} cd: {cd} cf: {cf}")
        print('')
        
uasyncio.run(main_loop())